-----------------------------------------------------------------------------------------------
-- Client Lua Script for GuildNation
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "bit32"
 
-----------------------------------------------------------------------------------------------
-- GuildNation Module Definition
-----------------------------------------------------------------------------------------------
local GuildNation = {} 
local cassPkg = nil
local userSettings = {}
local defaultSettings = {
	officerKey = ""
}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local DEBUG = false
local ALERT = true
local HAND_SHAKE_SEED = "TheseArentTheDroidsYoureLookingFor"
local CONFIG_CHANNEL = "GuildNationChannel="

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
-- copy defaults
for k,v in pairs(defaultSettings) do 
	userSettings[k] = v
end	

local function debug(msg)
	if DEBUG then
		Print(msg)
	end
end

local function alert(msg)
	if ALERT then
		Print(msg)
	end
end

function GuildNation:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.channelGuildNation = nil
	self.chatLogAddon = nil
	self.outQueue = self:CreateQueue()
	self.inQueue = self:CreateQueue()
	self.playerUnit = nil
	self.playerName = nil
	self.guildChannel = {}
	self.guildName = nil
	self.nation = nil
	self.nationKey = ""

	self.startupTimer = nil
	self.mainTimer = nil

    return o
end

function GuildNation:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		"CassPkg-1.1"
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- GuildNation OnLoad
-----------------------------------------------------------------------------------------------
function GuildNation:OnLoad()
	cassPkg = Apollo.GetPackage("CassPkg-1.1").tPackage
	self.xmlDoc = XmlDoc.CreateFromFile("GuildNation.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- GuildNation OnDocLoaded
-----------------------------------------------------------------------------------------------
function GuildNation:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "GuildNationForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
		self.wndMain:Show(false, true)
		Apollo.RegisterEventHandler("ChatMessage", "OnChatMessage", self)
		Apollo.RegisterEventHandler("GuildInfoMessage", "OnGuildMessageUpdated", self) -- notification that the guild info has changed
		Apollo.RegisterSlashCommand("nation", "OnGuildNationOn", self)

		self.startupTimer = ApolloTimer.Create(0.5, true, "OnStartupTimer", self)

	end
end

-- Save User Settings
function GuildNation:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	local tSave = {}
    for k,_ in pairs(defaultSettings) do 
		tSave[k] = userSettings[k] 
	end	
	return tSave
end

-- Restore Saved User Settings
function GuildNation:OnRestore(eType, t)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	for k,v in pairs(defaultSettings) do 
		if t[k] ~= nil then 
			userSettings[k] = t[k]
		else
			userSettings[k] = v
		end
	end	
end

-----------------------------------------------------------------------------------------------
-- GuildNation Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/nation"
function GuildNation:OnGuildNationOn()
	self.wndMain:FindChild("OfficerKeyEditBox"):SetText(userSettings.officerKey)
	self.wndMain:Invoke() -- show the window
end

-- on timer, this processes our outgoing messages
function GuildNation:OnStartupTimer()
	if self.playerUnit ~= nil and self.playerUnit:IsValid() then
		self.playerName = self.playerUnit:GetName()
		if self:GatherSettings() then
			self:JoinChannel()
			if self.channelGuildNation == nil then
				alert("Failed to join GuildNation channel or GuildNation configuration not setup.")
			end
		end
		self.startupTimer:Stop()
	else
		self.playerUnit = GameLib.GetPlayerUnit()
	end
end

function GuildNation:GatherSettings()
	self.nation = nil
	local guild = nil
	local guilds = GuildLib.GetGuilds()
	for _,g in pairs(guilds) do
		if g:GetType() == GuildLib.GuildType_Guild then
			guild = g
			break
		end
	end
	if guild == nil then
		alert("You are not part of a guild, you cannot use this addon")
		return false
	end

	self.guildName = guild:GetName()
	local info = guild:GetInfoMessage()
	if info == nil then
		alert("Your guild leadership hasn't set ANY configuration in Guild Additional Information")
		return false
	end

	local i = string.find(info,CONFIG_CHANNEL)
	if i == nil then
		alert("Your guild leadership hasn't set the correct CHANNEL configuration in Guild Additional Information")
		return false
	end			
	i = i + CONFIG_CHANNEL:len()
	local channelDetails = self:Split(info:sub(i),"=",1)
	local channelDetails = self:Split(channelDetails[1], ",", 2)
	if #channelDetails > 0 then
		self.nation = channelDetails[1]
		if #channelDetails > 1 then
			self.nationKey = channelDetails[2]
		end
	end

	return self.nation ~= nil
end

function GuildNation:JoinChannel()
	if self.nation ~= nil and self.nation:len() > 0 then
		local channelName = "GuildNationChannel_" .. self.nation
    	self.channelGuildNation = ICCommLib.JoinChannel(channelName, "OnGuildNationMessage", self)	
    	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Debug, "You have joined the " .. self.nation .. " GuildNation channel.", "")	
    	-- now figure out the guild channels
    	local channels = ChatSystemLib.GetChannels()
    	for i,v in ipairs(channels) do
    		if (v:GetType() == 15 or v:GetType() == 16) then
    			self.guildChannel[v:GetType()] = v
    		end
    	end
   		if self:GetChatAddon() then
			self.mainTimer = ApolloTimer.Create(1.0, true, "OnMainTimer", self)
   		end
   	else
		alert("Your guild leadership hasn't set the correct configuration in Guild Additional Information")
	end
end

function GuildNation:GetChatAddon() 
	if self.chatLogAddon == nil then
		local chatAddons = {"ChatLog", "ChatFixed", "ImprovedChatLog", "BetterChatLog"}
		for _,addon in ipairs(chatAddons) do
			self.chatLogAddon = Apollo.GetAddon(addon)
			if self.chatLogAddon ~= nil then
				debug("You are using the " .. addon .. " chat addon.")
				break
			end
		end
	end
	if self.chatLogAddon == nil then
		alert("You are currently using a Chat addon that is not compatible with GuildNation")
	end
	return self.chatLogAddon ~= nil 
end


function GuildNation:OnMainTimer()
	if self.playerName == nil or self.guildName == nil or self.channelGuildNation == nil then
		return
	end

	-- now with a joined channel and legitimately loaded player, we can process the queues
	-- send outgoing messages
	while self:GetQueueCount(self.outQueue) > 0 do
		local packet = self:DeQueueMessage(self.outQueue)
		if packet.tMessage.strSender == self.playerName then
			self:SendMessagePacket(packet)
		end
	end

	-- post incoming messages to chatlog
	if self.chatLogAddon ~= nil then
		while self:GetQueueCount(self.inQueue) > 0 do
			local packet = self:DeQueueMessage(self.inQueue)
			debug("Processing a input message")
			-- if this message is from a different guild in the nation, then we process it, otherwise it already should be in chatlog
			if packet.sourceGuild ~= self.guildName then
				local channel = self.guildChannel[packet.channelType]
				self.chatLogAddon:OnChatMessage(channel, packet.tMessage)
			else
				debug("Message received from your home guild, skipping")
			end
		end
	end
end

function GuildNation:OnGuildMessageUpdated(guildOwner)
	self.channelGuildNation = nil  -- self.channelGuildNation:Leave() no way to leave a channel ATM
	self.nation = nil
	self.startupTimer:Stop()
	self.startupTimer:Start()
end
-----------------------------------------------------------------------------------------------
-- Message & Packet Receipt Functions
-----------------------------------------------------------------------------------------------

function GuildNation:SendMessagePacket(packet)
	if self.channelGuildNation == nil then
		return
	end
	packet.sourceGuild = self.guildName
	local key = self.nationKey
	if packet.channelType == 16 then
		key = userSettings.officerKey
	end
	packet.handShake = self:encrypt(HAND_SHAKE_SEED,key)
	packet.tMessage = self:encrypt(cassPkg.SaveTable(packet.tMessage), key)
	debug("Sending Message on " .. self.nation)
	self.channelGuildNation:SendMessage(packet)

	-- for solo testing
--	self:OnGuildNationMessage(nil,packet)
end

function GuildNation:ReceiveMessagePacket(packet)
	local key = self.nationKey
	if packet.channelType == 16 then
		key = userSettings.officerKey
	end
	local handShakeTest = self:decrypt(packet.handShake,key)
	if handShakeTest ~= nil and handShakeTest == HAND_SHAKE_SEED then
		local decryptTable = self:decrypt(packet.tMessage,key)
		packet.tMessage = cassPkg.LoadTable(decryptTable)
		self:QueueMessage(self.inQueue,packet)
	else
		debug("Encryption handshake failed, message skipped. You don't have the correct key for this traffic: " .. key)
	end
end

-- record messages in chat, any originating from the player is re-broadcast on the channel to the nation
function GuildNation:OnChatMessage(channelCurrent, tMessage)
	local channelType = channelCurrent:GetType()
	if channelType == 15 or channelType == 16 then
		local packet = {}
		packet.channelType = channelType
		packet.tMessage = tMessage
		self:QueueMessage(self.outQueue,packet)
	end
end

-- here we receive a boadcast message, we should have already 
function GuildNation:OnGuildNationMessage(channel, packet)
	debug("Received a message on the GuildNation Channel")
	self:ReceiveMessagePacket(packet)
end

-----------------------------------------------------------------------------------------------
-- Utility functions for queue management
-----------------------------------------------------------------------------------------------
function GuildNation:CreateQueue ()
	return {first = 0, last = -1}
end

function GuildNation:QueueMessage(list, msg)
	local last = list.last + 1
	list.last = last
	list[last] = msg
end

function GuildNation:DeQueueMessage(list)
	local first = list.first
	if first > list.last then return nil end
	local msg = list[first]
	list[first] = nil        -- to allow garbage collection
	list.first = first + 1
	return msg
end

function GuildNation:GetQueueCount(list)
	if list.first > list.last then return 0 end
	return list.last + 1
end

function GuildNation:Split(str, delim, maxNb)
    -- Eliminate bad cases...
    if string.find(str, delim) == nil then
        return { str }
    end
    if maxNb == nil or maxNb < 1 then
        maxNb = 0    -- No limit
    end
    local result = {}
    local pat = "(.-)" .. delim .. "()"
    local nb = 0
    local lastPos
    for part, pos in string.gfind(str, pat) do
        nb = nb + 1
        result[nb] = part
        lastPos = pos
        if nb == maxNb then break end
    end
    -- Handle the last field
    if nb ~= maxNb then
        result[nb + 1] = string.sub(str, lastPos)
    end
    return result
end

function GuildNation:scramble (text,key,direction) 
	if text == nil or text:len() < 1 or key == nil or key:len() < 1 then
		return text
	end
	local base64 ='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
	local cipher = ""
	local primeKey = ""

	-- use only the base64 character of the key
	for i =1, key:len() do
		local c = key:sub(i,i)
		if string.find(base64,c) ~= nil then
			primeKey = primeKey .. c
		end
	end
	local keyLength = primeKey:len()

    for i = 1, text:len() do
    	local c = text:sub(i,i)  -- character we encrypt
    	local l = string.find(base64,c) -- location in base64
    	if l ~= nil then -- if this is not a legit base64 character, don't encrypt it
	    	local k = (i % keyLength) + 1
	    	local p = primeKey:sub(k,k) -- select a character in the primeKey
	    	local shift = string.find(base64,p)  -- locate primeKey character in base64
	   		local j = l + shift * direction
	   		if j > 64 then
	   			j = j - 64
	   		elseif j < 1 then
	   			j = j + 64
	   		end
	   		c = base64:sub(j,j)
    	end
        cipher = cipher .. c
    end
    return cipher

end

function GuildNation:encrypt(text,key)
	local cipher = self:scramble(text,key,1)
--	debug("encrypting: " .. text ..  " with key " .. key .. " -> " .. cipher)
	return cipher
end
 
function GuildNation:decrypt(text,key)
	local cipher = self:scramble(text,key,-1)
--	debug("decrypting: " .. text ..  " with key " .. key .. " -> " .. cipher)
	return cipher
end

-----------------------------------------------------------------------------------------------
-- GuildNationForm Functions
-----------------------------------------------------------------------------------------------
function GuildNation:OnClose( wndHandler, wndControl, eMouseButton )
	self.wndMain:Close() -- hide the window
end

function GuildNation:OnOfficerKeyChange( wndHandler, wndControl, strText )
	userSettings.officerKey = strText
end

-----------------------------------------------------------------------------------------------
-- GuildNation Instance
-----------------------------------------------------------------------------------------------
local GuildNationInst = GuildNation:new()
GuildNationInst:Init()
